@@ -4,7 +4,7 @@
 
 ;; Author: Jim Porter
 ;; URL: https://github.com/jimporter/urgrep
-;; Version: 0.4.1
+;; Version: 0.5.0
 ;; Keywords: grep, search
 ;; Package-Requires: ((emacs "27.1") (compat "29.1.0.1") (project "0.3.0"))
 
@@ -197,7 +197,7 @@ for MS shells."
 This expands tildes and removes any remote host identifiers or quoting."
   (when-let ((remote (file-remote-p file)))
     (unless (equal remote (file-remote-p default-directory))
-      (error "remote file doesn't match host for `default-directory'"))
+      (error "Remote file doesn't match host for `default-directory'"))
     (setq file (file-local-name file)))
   (setq file (file-name-unquote file))
   (if (not (string-prefix-p "~" file))
@@ -261,7 +261,7 @@ properties defined in the `urgrep-tools' entry for TOOL."
   "Add EXTRA-OPTS to the specified GREP template and return it.
 This function changes match data."
   (unless (string-match "<C>" template)
-    (error "grep template should have a <C> placeholder"))
+    (error "Grep template should have a <C> placeholder"))
   ;; Locally add options to the template that grep.el isn't aware of.
   (replace-match (concat "<C> " (urgrep--shell-join extra-opts))
                  t t template))
@@ -631,6 +631,18 @@ it up in `urgrep-tools'.  Otherwise, return TOOL as-is."
           ((and (eq syntax 'pcre) (memq 'extended tool-syntaxes)) 'ere)
           (t (car tool-syntaxes)))))
 
+(defun urgrep--case-fold-p (case-fold query regexp-syntax)
+  "Return whether to enable case folding.
+If CASE-FOLD is `inherit' or `smart', guess based on the QUERY
+and REGEXP-SYNTAX.  Otherwise, just return CASE-FOLD."
+  (when (eq case-fold 'inherit)
+    (setq case-fold (if case-fold-search 'smart nil)))
+  (when (eq case-fold 'smart)
+    (when regexp-syntax
+      (setq query (urgrep--convert-regexp query regexp-syntax 'bre)))
+    (setq case-fold (isearch-no-upper-case-p query regexp-syntax)))
+  case-fold)
+
 ;;;###autoload
 (cl-defun urgrep-command (query &key tool regexp (case-fold 'inherit) hidden
                                 file-wildcard (root ".") (group t) (context 0)
@@ -671,19 +683,15 @@ respectively).
 
 COLOR: non-nil (the default) if the output should use color."
   (with-connection-local-variables
-   (let* ((regexp-syntax (if (eq regexp t) urgrep-regexp-syntax regexp))
+   (let* ((tool (or (urgrep-get-tool tool)
+                    (error "Unknown tool %s" tool)))
+          (regexp-syntax (if (eq regexp t) urgrep-regexp-syntax regexp))
+          (tool-re-syntax (urgrep--get-best-syntax regexp-syntax tool))
+          (case-fold (urgrep--case-fold-p case-fold query regexp-syntax))
           (file-wildcard (ensure-list file-wildcard))
           (root (mapcar #'urgrep--safe-file-name (ensure-list root)))
-          (tool (or (urgrep-get-tool tool)
-                    (error "unknown tool %s" tool)))
-          (tool-re-syntax (urgrep--get-best-syntax regexp-syntax tool))
           (query (urgrep--convert-regexp query regexp-syntax tool-re-syntax))
           (cmd-fun (urgrep--get-prop 'command-function tool)))
-     ;; Determine whether to search case-sensitively or not.
-     (when (eq case-fold 'inherit)
-       (setq case-fold (if case-fold-search 'smart nil)))
-     (when (eq case-fold 'smart)
-       (setq case-fold (isearch-no-upper-case-p query regexp-syntax)))
      ;; Build the command arguments.
      (if cmd-fun
          (funcall cmd-fun query :tool tool :regexp tool-re-syntax
@@ -743,6 +751,77 @@ If EDIT-COMMAND is non-nil, the search can be edited."
                     query)))
     (urgrep--start command query urgrep-current-tool)))
 
+(defmacro urgrep-adjust-query (prop &rest body)
+  "Adjust the current search query by updating PROP.
+This gets PROP from `urgrep-current-query' and let-binds it to a
+variable of the same name, excluding the leading colon (so,
+e.g. `:context' becomes `context'), and then evaluates BODY.
+Finally, this sets PROP to BODY's return value and re-runs the
+search."
+  (declare (indent 1))
+  (unless (keywordp prop)
+    (error "Invalid PROP %S" prop))
+  (let ((prop-variable (intern (substring (symbol-name prop) 1))))
+    `(if (not (listp urgrep-current-query))
+         (error "Unable to determine query properties")
+       (let* ((new-value
+               (let ((,prop-variable (plist-get (cdr urgrep-current-query)
+                                                ,prop)))
+                 ,@body))
+              (new-query (cons (car urgrep-current-query)
+                               (plist-put (cdr urgrep-current-query)
+                                          ,prop new-value))))
+         (urgrep--start (apply #'urgrep-command new-query) new-query
+                        urgrep-current-tool)))))
+
+(defun urgrep-toggle-current-case-fold ()
+  "Toggle case folding for the current query."
+  (interactive)
+  (urgrep-adjust-query :case-fold
+    (let ((query (car urgrep-current-query))
+          (regexp-syntax (plist-get (cdr urgrep-current-query) :regexp-syntax)))
+      (not (urgrep--case-fold-p case-fold query regexp-syntax)))))
+
+(defun urgrep-toggle-current-search-hidden-files ()
+  "Toggle searching of hidden files for the current query."
+  (interactive)
+  (urgrep-adjust-query :hidden
+    (not hidden)))
+
+(defun urgrep-expand-current-context (&optional lines)
+  "Expand the context of the current query by LINES.
+Interactively, LINES is the numeric prefix argument.  If LINES is
+nil, expand the context by one line."
+  (interactive "p")
+  (setq lines (or lines 1))
+  (urgrep-adjust-query :context
+    (if (consp context)
+        (cons (max (+ (car context) lines) 0)
+              (max (+ (cdr context) lines) 0))
+      (max (+ (or context 0) lines) 0))))
+
+(defun urgrep-expand-current-before-context (&optional lines)
+  "Expand the leading context of the current query by LINES.
+Interactively, LINES is the numeric prefix argument.If LINES is
+nil, expand the leading context by one line."
+  (interactive "p")
+  (setq lines (or lines 1))
+  (urgrep-adjust-query :context
+    (let ((context (if (consp context) context
+                     (cons (or context 0) (or context 0)))))
+      (cons (max (+ (car context) lines) 0) (cdr context)))))
+
+(defun urgrep-expand-current-after-context (&optional lines)
+  "Expand the trailing context of the current query by LINES.
+Interactively, LINES is the numeric prefix argument.  If LINES is
+nil, expand the trailing context by one line."
+  (interactive "p")
+  (setq lines (or lines 1))
+  (urgrep-adjust-query :context
+    (let ((context (if (consp context) context
+                     (cons (or context 0) (or context 0)))))
+      (cons (car context) (max (+ (cdr context) lines) 0)))))
+
 (defvar-keymap urgrep-mode-map
   ;; Don't inherit from `compilation-minor-mode-map', because that introduces a
   ;; menu bar item we don't want.
@@ -763,7 +842,12 @@ If EDIT-COMMAND is non-nil, the search can be edited."
   "}"             #'compilation-next-file
   "TAB"           #'compilation-next-error
   "<backtab>"     #'compilation-previous-error
-  "g"             #'urgrep-search-again)
+  "g"             #'urgrep-search-again
+  "c"             #'urgrep-toggle-current-case-fold
+  "H"             #'urgrep-toggle-current-search-hidden-files
+  "C"             #'urgrep-expand-current-context
+  "B"             #'urgrep-expand-current-before-context
+  "A"             #'urgrep-expand-current-after-context)
 
 (easy-menu-define urgrep-menu-map urgrep-mode-map
   "Menu for urgrep buffers."
@@ -1268,15 +1352,24 @@ future searches."
 
 (defvar-keymap urgrep-minibuffer-map
   :parent minibuffer-local-map
+  "M-s h" #'urgrep-minibuffer-help
   "M-s r" #'urgrep-toggle-regexp
   "M-s c" #'urgrep-toggle-case-fold
   "M-s f" #'urgrep-set-file-wildcards
-  "M-s h" #'urgrep-toggle-search-hidden-files
+  "M-s H" #'urgrep-toggle-search-hidden-files
   "M-s C" #'urgrep-set-context
   "M-s B" #'urgrep-set-before-context
   "M-s A" #'urgrep-set-after-context
   "M-s t" #'urgrep-set-tool)
 
+(defun urgrep-minibuffer-help ()
+  "Describe key bindings for the Urgrep minibuffer."
+  (interactive)
+  ;; `describe-keymap' is only available in Emacs 28+.
+  (if (fboundp 'describe-keymap)
+      (describe-keymap urgrep-minibuffer-map)
+    (describe-bindings)))
+
 (cl-defun urgrep--read-query (initial &key tool (regexp urgrep-search-regexp)
                                       (case-fold urgrep-case-fold)
                                       (hidden urgrep-search-hidden-files)
